"""
Iterate over priority domains and send data to server
"""
import argparse
import json
import logging
from time import sleep
import requests
import sys
from whoare.whoare import WhoAre
from whoare import __version__

logger = logging.getLogger(__name__)


class WhoAreShare:
    def __init__(self, get_domains_url, post_url, token, torify=True, pause_between_calls=41, from_path=None):
        self.torify = torify  # use local IP and also torify
        self.post_url = post_url  # destination URL to share data (will be processed outside)
        self.token = token
        self.get_domains_url = get_domains_url  # URL to get domains from
        self.pause_between_calls = pause_between_calls
        self.from_path = from_path

    def run(self):
        """ get domains and _whois_ them """
        if self.from_path is not None:
            self.run_from_path(self.from_path)
        else:
            self.run_from_priority()

    def run_from_priority(self):
        """ get priority domains from API """
        logger.info('Start runing from priority')
        while True:
            
            domain = self.get_one()
            self.load_one(domain, torify=False)

            # if torify start a second queue
            if self.torify:
                self.load_one(domain, torify=True)
            
            sleep(self.pause_between_calls)
    
    def run_from_path(self, path):
        """ open a file and update those domains """
        f = open(path)
        domain_data = f.read()
        f.close()

        domain_list = domain_data.split('\n')

        c = 0
        while True:
            
            domain = domain_list[c]
            self.load_one(domain, torify=False)

            if self.torify:
                c += 1
                if c >= len(domain_list): break
                domain = domain_list[c]
                self.load_one(domain, torify=True)
            
            sleep(self.pause_between_calls)
            c += 1
            if c >= len(domain_list): break

    def load_one(self, domain, torify):
        """ analyze and push one domain """
        logger.info(f'Domain {domain} {torify}')

        wa = WhoAre()
        try:
            wa.load(domain, torify=torify)
        except:
            pass
        else:
            self.post_one(wa)

    def get_one(self):
        """ get the next priority from API """
        logger.info('Getting one')
        headers = {'Authorization': f'Token {self.token}'}
        response = requests.get(self.get_domains_url, headers=headers)
        if response.status_code != 200:
            raise ValueError(f'Error GET status {response.status_code}: {response.text}')
        
        try:
            jresponse = response.json()
        except Exception:
            print(f'ERROR parsing {response.text}')
            raise
        
        logger.info(f' - Get {jresponse}')
        
        return jresponse[0]['domain']
    
    def post_one(self, wa):
        """ post results to server """
        logger.info(f'POST {wa}')
        headers = {'Authorization': f'Token {self.token}'}
        data = wa.as_dict()
        data['whoare_version'] = __version__
        logger.info(f'POSTing {data}')
        str_data = json.dumps(data)
        final = {'domain': str_data}
        response = requests.post(self.post_url, data=final, headers=headers)
        jresponse = response.json()
        logger.info(f' - POST {jresponse}')
        return jresponse['ok']

def main():
        
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # base_domain = 'http://localhost:8000'
    base_domain = 'https://nic.opendatacordoba.org'
    default_get = f'{base_domain}/api/v1/dominios/next-priority/'
    default_post = f'{base_domain}/api/v1/dominios/dominio/update_from_whoare/'
    
    parser = argparse.ArgumentParser(prog='whoare-share')
    parser.add_argument('--get', nargs='?', help='URL to get domains from', type=str, default=default_get)
    parser.add_argument('--post', nargs='?', help='URL to post results to', type=str, default=default_post)
    parser.add_argument('--token', nargs='?', help='Token to use as Header Autorization', type=str, required=True)
    parser.add_argument('--torify', nargs='?', type=bool, default=True, help='Use torify for WhoIs command')
    parser.add_argument('--pause', nargs='?', help='Pause between calls', default=41, type=int)
    parser.add_argument('--from_path', nargs='?', help='If not used we will get priorities from API. This is usted for new-domain lists', type=str)
    
    args = parser.parse_args()

    was = WhoAreShare(
        get_domains_url=args.get,
        post_url=args.post,
        token=args.token,
        torify=args.torify,
        pause_between_calls=args.pause,
        from_path=args.from_path
    )

    was.run()


class WhoIsCommandError(Exception):
    pass


class ZoneNotFoundError(Exception):
    pass


class TooManyQueriesError(Exception):
    """ Registrar ask to stop querying """
    pass


class ServiceUnavailableError(Exception):
    """ Registrar ask to stop querying """
    pass

class UnknownError(Exception):
    """ Registrar asks to stop querying """
    pass

class UnexpectedParseError(Exception):
    """ Registrar asks to stop querying """
    pass

class UnexpectedDomainError(Exception):
    """ Registrar asks to stop querying """
    pass


from datetime import datetime
import logging
import subprocess

from whoare.base import Domain, subclasses, Registrant, DNS
from whoare.exceptions import ZoneNotFoundError, WhoIsCommandError

logger = logging.getLogger(__name__)


class WhoAre:

    def __init__(self):
        self.child = None  # the object for the zone
        self.domain = None  # Domain Object
        self.registrant = None  # Registrant Objects
        self.dnss = []  # all DNSs objects
        self.raw_data = None  # whois response

    def get_raw(self, domain, host=None, torify=False):
        if host:
            command = ['whois', f'-h {host}', domain]
        else:
            command = ['whois', domain]
        
        if torify:
            command = ['torify'] + command

        p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        r = p.communicate()[0]
        raw = r.decode()
        
        if p.returncode != 0:
            error = f'WhoIs error {p.returncode} {raw}'
            logger.error(error)
            raise WhoIsCommandError(error)
        
        self.raw_data = raw
        return raw

    def load(self, domain, host=None, mock_from_txt_file=None, torify=False):
        """ load domain data. 
                domain is DOMAIN.ZONE (never use subdomain like www or others)
                host could be "whois.nic.ar" of argentina (optional) 
            Return a dict with parsed data and fill class properties 
                mock_from_txt_file could be used to a path with exact whois results """
        
        logger.info(f'Load {domain} {host}')
        
        domain_name, zone = self.detect_zone(domain)
        zone_class = self.detect_subclass(zone)
        self.child = zone_class()
        self.domain = Domain(domain_name, zone)

        logger.info(f'Zone Class {zone_class} {domain_name} {zone}')
        
        domain = f'{domain_name}.{zone}'

        if mock_from_txt_file is not None:
            f = open(mock_from_txt_file)
            raw = f.read()
            f.close()
        else:
            raw = self.get_raw(domain, host, torify)
            
        return self.load_from_raw(raw)

    def load_from_raw(self, raw):

        if self.child.is_free(raw):
            return None

        # raise any errors
        self.child.check_errors(raw)

        self.domain.is_free = False
        # parse raw data
        self.child.parse(raw, self)
        
    def detect_zone(self, domain):
        logger.info(f'Detect zone {domain}')
        
        domain = domain.lower().strip()

        parts = domain.split('.')
        if parts[0].startswith('https://'):
            parts[0].replace('https://', '')
        elif parts[0].startswith('http://'):
            parts[0].replace('http://', '')

        domain = parts[0]
        zone = '.'.join(parts[1:])

        return domain, zone
    
    def detect_subclass(self, zone):
        
        logger.info(f'Detecting subclass for {zone} at {subclasses}')
        
        for cls in subclasses:
            logger.info(f'Searching zones for {cls} {cls.zones()}')
            if zone in cls.zones():
                return cls

        error = f'Zone not covered "{zone}"'
        logger.error(error)
        raise ZoneNotFoundError(error)

    def as_dict(self):
        """ build a nice dict """
        res = {
                "domain": {
                    "base_name": self.domain.base_name,
                    "zone": self.domain.zone,
                    "is_free": self.domain.is_free
                    }
                }
        if not self.domain.is_free:
            res["domain"]["registered"] = self.domain.registered.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["domain"]["changed"] = self.domain.changed.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["domain"]["expire"] = self.domain.expire.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["registrant"] = {
                "name": self.registrant.name,
                "legal_uid": self.registrant.legal_uid,
                "created": self.registrant.created.strftime('%Y-%m-%d %H:%M:%S.%f %z'),
                "changed": self.registrant.changed.strftime('%Y-%m-%d %H:%M:%S.%f %z')
                }

            res["dnss"] = [dns.name for dns in self.dnss]
            
        logger.info(f'as dict {res}')
        return res

    def from_dict(self, data):
        """ load object from dict """

        self.domain = Domain(data['domain']['base_name'], data['domain']['zone'])
        self.registrant = None
        self.dnss = []
        self.domain.is_free = data['domain']['is_free']
        if self.domain.is_free:
            return
        
        self.domain.registered = datetime.strptime(data["domain"]["registered"], '%Y-%m-%d %H:%M:%S.%f %z')
        self.domain.changed = datetime.strptime(data["domain"]["changed"], '%Y-%m-%d %H:%M:%S.%f %z')
        self.domain.expire = datetime.strptime(data["domain"]["expire"] , '%Y-%m-%d %H:%M:%S.%f %z')
        
        self.registrant = Registrant(name=data['registrant']['name'], legal_uid=data['registrant']['legal_uid'])
        self.registrant.created = datetime.strptime(data['registrant']['created'], '%Y-%m-%d %H:%M:%S.%f %z')
        self.registrant.changed = datetime.strptime(data['registrant']['changed'], '%Y-%m-%d %H:%M:%S.%f %z')

        for ns in data["dnss"]:
            self.dnss.append(DNS(name=ns))

    def __eq__(self, wa2):
        diffs = []
        if self.domain.base_name != wa2.domain.base_name: 
            diffs.append(f'domain.base_name {self.domain.base_name} != {wa2.domain.base_name}')
        if self.domain.zone != wa2.domain.zone:
            diffs.append(f'domain.zone {self.domain.zone} != {wa2.domain.zone}')
        if self.domain.is_free != wa2.domain.is_free:
            diffs.append('domain.is_free')
        if self.domain.registered != wa2.domain.registered:
            diffs.append(f'domain.registered {self.domain.registered} != {wa2.domain.registered}')
        if self.domain.changed != wa2.domain.changed:
            diffs.append(f'domain.changed {self.domain.changed} != {wa2.domain.changed}')
        if self.domain.expire != wa2.domain.expire:
            diffs.append(f'domain.expire {self.domain.expire} != {wa2.domain.expire}')
        
        if (self.registrant is None) != (wa2.registrant is None):
            diffs.append(f'registrant {self.registrant} != {wa2.registrant}')
        if self.registrant is not None and wa2.registrant is not None:
            if self.registrant.name != wa2.registrant.name:
                diffs.append(f'registrant.name {self.registrant.name} != {wa2.registrant.name}')
            if self.registrant.legal_uid != wa2.registrant.legal_uid:
                diffs.append(f'registrant.legal_uid {self.registrant.legal_uid} != {wa2.registrant.legal_uid}')
            if self.registrant.created != wa2.registrant.created:
                diffs.append(f'registrant.created {self.registrant.created} != {wa2.registrant.created}')
            if self.registrant.changed != wa2.registrant.changed:
                diffs.append(f'registrant.changed {self.registrant.changed} != {wa2.registrant.changed}')
        
        if len(self.dnss) != len(wa2.dnss): 
            diffs.append('Len DNSs')
        else:
            c = 0
            for ns in self.dnss:
                if ns.name != wa2.dnss[c].name:
                    diffs.append(f'DNS {c}. {ns.name} != {wa2.dnss[c].name}')
                c += 1

        logger.info(f'DIFFs {diffs}')
        return len(diffs) == 0

    def as_plain_dict(self):
        """ build a nice dict """
        res = {
            "domain_base_name": self.domain.base_name,
            "domain_zone": self.domain.zone,
            "domain_is_free": self.domain.is_free
            }

        if not self.domain.is_free:
            res["domain_registered"] = self.domain.registered.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["domain_changed"] = self.domain.changed.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["domain_expire"] = self.domain.expire.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["registrant_name"] = self.registrant.name
            res["registrant_legal_uid"] = self.registrant.legal_uid
            res["registrant_created"] = self.registrant.created.strftime('%Y-%m-%d %H:%M:%S.%f %z')
            res["registrant_changed"] = self.registrant.changed.strftime('%Y-%m-%d %H:%M:%S.%f %z')

            c = 1
            for dns in self.dnss:
                res[f"dns{c}"] = dns.name
                c += 1
            
        logger.info(f'as plain dict {res}')
        return res

    def from_plain_dict(self, data):
        """ load object from dict """

        self.domain = Domain(data['domain_base_name'], data['domain_zone'])
        self.registrant = None
        self.dnss = []
        self.domain.is_free = data['domain_is_free']
        if self.domain.is_free:
            return
        
        self.domain.registered = datetime.strptime(data["domain_registered"], '%Y-%m-%d %H:%M:%S.%f %z')
        self.domain.changed = datetime.strptime(data["domain_changed"], '%Y-%m-%d %H:%M:%S.%f %z')
        self.domain.expire = datetime.strptime(data["domain_expire"] , '%Y-%m-%d %H:%M:%S.%f %z')
        
        self.registrant = Registrant(name=data['registrant_name'], legal_uid=data['registrant_legal_uid'])
        self.registrant.created = datetime.strptime(data['registrant_created'], '%Y-%m-%d %H:%M:%S.%f %z')
        self.registrant.changed = datetime.strptime(data['registrant_changed'], '%Y-%m-%d %H:%M:%S.%f %z')

        for c in range(1, 20):
            if f"dns{c}" in data:
                self.dnss.append(DNS(name=data[f"dns{c}"]))
            else:
                break



import logging
import subprocess
from whoare.zone_parsers.ar.who import WhoAr

logger = logging.getLogger(__name__)

subclasses = [
    WhoAr
]

class Domain:
    
    def __init__(self, base_name, zone):
        logger.info(f'Domain object created {base_name}, {zone}')
        self.base_name = base_name.lower().strip()
        self.zone = zone.lower().strip()
        self.registered = None
        self.changed = None
        self.expire = None
        self.is_free = True  # change is registered

    def __str__(self):
        if self.is_free:
            return f'Domain {self.full_name()} FREE'
        return f'Domain {self.base_name} at {self.zone}. Registered at {self.registered}'
    
    def full_name(self):
        return f'{self.base_name}.{self.zone}'

class Registrant:
    
    def __init__(self, name, legal_uid):
        logger.info(f'Registrant object created {name}, {legal_uid}')
        self.name = name.lower().strip()
        self.legal_uid = legal_uid.lower().strip()
        self.created = None
        self.changed = None

    def __str__(self):
        return f'Registrant {self.name} UID {self.legal_uid}'
    

class DNS:
    
    def __init__(self, name):
        logger.info(f'DNS object created {name}')
        self.name = name.lower().strip()
    
    def __str__(self):
        return f'DNS {self.name}'
    


    # def is_free(self, raw):
    #     """ read raw response and detect if the domain is free """
    #     raise NotImplementedError

    # def check_errors(self, raw):
    #     """ read raw response and detect if there are errors """
    #     raise NotImplementedError
        
    # def parse(self, raw):
    #     """ read raw response and fill all data """
    #     raise NotImplementedError



package_name = 'whoare'
__version__ = '0.1.31'

def require_update_pypi():
    import requests
    response = requests.get(f'https://pypi.org/pypi/{package_name}/json')
    data = response.json()
    pypi_version = data['info']['version']

    return pypi_version != __version__
